# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import subprocess
import subprocess as sp
import tempfile
from pathlib import Path
from typing import Optional
from zipfile import ZipFile

from kicadcliwrapper.generated.kicad_cli import kicad_cli as k

logger = logging.getLogger(__name__)


class KicadCliExportError(Exception):
    pass


def _export(cmd):
    return k(k.pcb(k.pcb.export(cmd))).exec()


def githash_layout(layout: Path, out: Path) -> Path:
    # Get current git hash
    try:
        git_hash = (
            subprocess.check_output(
                ["git", "rev-parse", "--short", "HEAD"], cwd=layout.parent
            )
            .decode("ascii")
            .strip()
        )
    except (subprocess.CalledProcessError, FileNotFoundError):
        logger.warning("Could not get git hash, using 'unknown'")
        git_hash = "unknown"

    # Read and substitute git hash
    layout_text = layout.read_text(encoding="utf-8")
    layout_text = layout_text.replace("{{GITHASH}}", git_hash)

    # Write modified layout to temp file
    out.write_text(layout_text, encoding="utf-8")

    return out


def export_step(
    pcb_file: Path, step_file: Path, project_dir: Optional[Path] = None
) -> None:
    """
    3D PCBA STEP file export using the kicad-cli
    """

    # If project_dir is provided, set KIPRJMOD to ensure 3D model paths resolve
    cmd_args = {
        "INPUT_FILE": str(pcb_file),
        "force": True,
        "no_dnp": True,
        "subst_models": True,
        "output": str(step_file.absolute()),
    }

    if project_dir:
        cmd_args["define_var"] = f"KIPRJMOD={project_dir.absolute()}"

    try:
        _export(k.pcb.export.step(**cmd_args))
    except sp.CalledProcessError as e:
        raise KicadCliExportError("Failed to export step file") from e


def export_dxf(pcb_file: Path, dxf_file: Path) -> None:
    """
    PCB outline export using the kicad-cli
    """

    try:
        _export(
            k.pcb.export.dxf(
                INPUT_FILE=str(pcb_file),
                exclude_refdes=True,
                exclude_value=True,
                output_units="mm",
                layers="Edge.Cuts",
                output=str(dxf_file),
            )
        )
    except sp.CalledProcessError as e:
        raise KicadCliExportError("Failed to export dxf file") from e


def export_glb(
    pcb_file: Path, glb_file: Path, project_dir: Optional[Path] = None
) -> None:
    """
    3D PCBA GLB file export using the kicad-cli
    """

    # If project_dir is provided, set KIPRJMOD to ensure 3D model paths resolve
    cmd_args = {
        "INPUT_FILE": str(pcb_file),
        "force": True,
        "include_tracks": True,
        "include_zones": True,
        "grid_origin": True,
        "subst_models": True,
        "no_dnp": True,
        "cut_vias_in_body": True,
        "include_pads": True,
        "include_soldermask": True,
        "include_silkscreen": True,
        "output": str(glb_file.absolute()),
    }

    if project_dir:
        cmd_args["define_var"] = f"KIPRJMOD={project_dir.absolute()}"

    try:
        _export(k.pcb.export.glb(**cmd_args))
    except sp.CalledProcessError as e:
        raise KicadCliExportError("Failed to export glb file") from e


def export_svg(pcb_file: Path, svg_file: Path, flip_board: bool = False) -> None:
    """
    2D PCBA SVG file export using the kicad-cli
    """

    layers = "F.Cu,F.Paste,F.SilkS,F.Mask,Edge.Cuts"
    if flip_board:
        layers = layers.replace("F.", "B.")

    try:
        _export(
            k.pcb.export.svg(
                INPUT_FILE=str(pcb_file),
                layers=layers,
                page_size_mode="2",
                exclude_drawing_sheet=True,
                output=str(svg_file),
            )
        )
    except sp.CalledProcessError as e:
        raise KicadCliExportError("Failed to export svg file") from e


def export_gerber(pcb_file: Path, gerber_zip_file: Path) -> None:
    """
    Gerber export using the kicad-cli
    """

    logger.info(f"Exporting gerber files to {gerber_zip_file}")
    gerber_dir = gerber_zip_file.parent
    gerber_dir.mkdir(parents=True, exist_ok=True)

    # Create a temporary folder to export the gerber and drill files to
    with tempfile.TemporaryDirectory(dir=gerber_dir) as temp_dir:
        out_dir = temp_dir if temp_dir.endswith("/") else f"{temp_dir}/"

        try:
            _export(
                k.pcb.export.gerbers(
                    INPUT_FILE=str(pcb_file),
                    output=out_dir,
                )
            )
        except sp.CalledProcessError as e:
            raise KicadCliExportError("Failed to export gerber files") from e

        try:
            _export(
                k.pcb.export.drill(
                    INPUT_FILE=str(pcb_file),
                    format="excellon",
                    excellon_separate_th=True,
                    generate_map=True,
                    map_format="gerberx2",
                    output=out_dir,
                )
            )

        except sp.CalledProcessError as e:
            raise KicadCliExportError("Failed to export drill files") from e

        # Zip the gerber files
        with ZipFile(gerber_zip_file, "w") as zipf:
            for file in Path(temp_dir).iterdir():
                if file.is_file():
                    zipf.write(file, arcname=file.name)


def export_pick_and_place(pcb_file: Path, pick_and_place_file: Path) -> None:
    """
    Pick and place export using the kicad-cli
    """

    try:
        _export(
            k.pcb.export.pos(
                INPUT_FILE=str(pcb_file),
                side="both",
                format="csv",
                units="mm",
                exclude_dnp=True,
                output=str(pick_and_place_file),
            )
        )
    except sp.CalledProcessError as e:
        raise KicadCliExportError("Failed to export pick and place file") from e



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from enum import Enum, auto

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P, quantity
from faebryk.libs.util import assert_once


class LDO(Module):
    @assert_once
    def enable_output(self):
        self.enable.set(True)

    class OutputType(Enum):
        FIXED = auto()
        ADJUSTABLE = auto()

    class OutputPolarity(Enum):
        POSITIVE = auto()
        NEGATIVE = auto()

    max_input_voltage = L.p_field(
        units=P.V,
        likely_constrained=True,
        soft_set=L.Range(1 * P.V, 100 * P.V),
    )
    output_voltage = L.p_field(
        units=P.V,
        likely_constrained=True,
        soft_set=L.Range(1 * P.V, 100 * P.V),
    )
    quiescent_current = L.p_field(
        units=P.A,
        likely_constrained=True,
        soft_set=L.Range(1 * P.mA, 100 * P.mA),
    )
    dropout_voltage = L.p_field(
        units=P.V,
        likely_constrained=True,
        soft_set=L.Range(1 * P.mV, 100 * P.mV),
    )
    ripple_rejection_ratio = L.p_field(
        units=P.dB,
        likely_constrained=True,
        soft_set=L.Range(quantity(1, P.dB), quantity(100, P.dB)),
    )
    output_polarity = L.p_field(
        domain=L.Domains.ENUM(OutputPolarity),
    )
    output_type = L.p_field(
        domain=L.Domains.ENUM(OutputType),
    )
    output_current = L.p_field(
        units=P.A,
        likely_constrained=True,
        soft_set=L.Range(1 * P.mA, 100 * P.mA),
    )
    enable: F.EnablePin
    power_in: F.ElectricPower
    power_out = L.d_field(lambda: F.ElectricPower().make_source())

    # @L.rt_field
    # def pickable(self) -> F.is_pickable_by_type:
    #     return F.is_pickable_by_type(
    #         F.is_pickable_by_type.Type.LDO,
    #         {
    #             "max_input_voltage": self.max_input_voltage,
    #             "output_voltage": self.output_voltage,
    #             "quiescent_current": self.quiescent_current,
    #             "dropout_voltage": self.dropout_voltage,
    #             # TODO: add support in backend
    #             # "ripple_rejection_ratio": self.ripple_rejection_ratio,
    #             "output_polarity": self.output_polarity,
    #             "output_type": self.output_type,
    #             "output_current": self.output_current,
    #         },
    #     )

    def __preinit__(self):
        self.max_input_voltage.constrain_ge(self.power_in.voltage)
        self.power_out.voltage.constrain_subset(self.output_voltage)

        self.enable.enable.reference.connect(self.power_in)
        # TODO: should be implemented differently (see below)
        # if self.output_polarity == self.OutputPolarity.NEGATIVE:
        #    self.power_in.hv.connect(self.power_out.hv)
        # else:
        #    self.power_in.lv.connect(self.power_out.lv)

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self, gnd_only=True)
        )

    @L.rt_field
    def decoupled(self):
        return F.can_be_decoupled_rails(self.power_in, self.power_out)

    @L.rt_field
    def can_bridge(self):
        return F.can_bridge_defined(self.power_in, self.power_out)

    @L.rt_field
    def simple_value_representation(self):
        S = F.has_simple_value_representation_based_on_params_chain.Spec
        return F.has_simple_value_representation_based_on_params_chain(
            S(self.output_voltage, tolerance=True),
            S(self.output_current),
            S(self.ripple_rejection_ratio),
            S(self.dropout_voltage),
            S(self.max_input_voltage, prefix="Vin max"),
            S(self.quiescent_current, prefix="Iq"),
            prefix="LDO",
        )

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )

    @L.rt_field
    def pin_association_heuristic(self):
        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                self.power_in.hv: ["Vin", "Vi", "in"],
                self.power_out.hv: ["Vout", "Vo", "out", "output"],
                self.power_in.lv: ["GND", "V-", "ADJ/GND"],
                self.enable.enable.line: ["EN", "Enable"],
            },
            accept_prefix=False,
            case_sensitive=False,
        )



import logging

import typer

from atopile.mcp.mcp_server import run_mcp
from atopile.telemetry import capture

logger = logging.getLogger(__name__)


mcp_app = typer.Typer(rich_markup_mode="rich")


@mcp_app.command()
@capture("cli:mcp_start", "cli:mcp_end")
def start(http: bool = False, debug: bool = False):
    """Start the MCP server."""
    run_mcp(http, debug)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import faebryk.library._F as F
from faebryk.exporters.pcb.layout.layout import Layout


class has_pcb_layout_defined(F.has_pcb_layout.impl()):
    def __init__(self, layout: Layout) -> None:
        super().__init__()
        self.layout = layout

    def apply(self):
        node = self.obj
        return self.layout.apply(node)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from types import UnionType
from typing import TYPE_CHECKING, overload

from faebryk.core.cpp import Graph
from faebryk.core.node import Node

if TYPE_CHECKING:
    from faebryk.core.trait import Trait

logger = logging.getLogger(__name__)


# TODO move these to C++
# just here for making refactoring easier for the moment
# a bit weird typecheck
class GraphFunctions:
    # Make all kinds of graph filtering functions so we can optimize them in the future
    # Avoid letting user query all graph nodes always because quickly very slow
    def __init__(self, *graph: Graph):
        self.graph = graph

    def node_projection(self) -> list["Node"]:
        return list(self.nodes_of_type(Node))

    def nodes_with_trait[T: "Trait"](self, trait: type[T]) -> list[tuple["Node", T]]:
        return [
            (n, n.get_trait(trait))
            for n in self.node_projection()
            if n.has_trait(trait)
        ]

    # TODO: Waiting for python to add support for type mapping
    def nodes_with_traits[*Ts](
        self, traits: tuple[*Ts]
    ):  # -> list[tuple[Node, tuple[*Ts]]]:
        return [
            (n, tuple(n.get_trait(trait) for trait in traits))  # type: ignore
            for n in self.node_projection()
            if all(n.has_trait(trait) for trait in traits)  # type: ignore
        ]

    def nodes_of_type[T: "Node"](self, t: type[T]) -> set[T]:
        return {n for g in self.graph for n in g.node_projection() if isinstance(n, t)}

    @overload
    def nodes_of_types(self, t: tuple[type["Node"], ...]) -> set["Node"]: ...
    @overload
    def nodes_of_types(self, t: UnionType) -> set["Node"]: ...

    def nodes_of_types(self, t):  # type: ignore TODO
        return {n for g in self.graph for n in g.node_projection() if isinstance(n, t)}



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import faebryk.library._F as F
from faebryk.libs.library import L
from faebryk.libs.units import P, Quantity
from faebryk.libs.util import times


class DIP(F.Footprint):
    def __init__(self, pin_cnt: int, spacing: Quantity, long_pads: bool) -> None:
        super().__init__()

        self.spacing = spacing
        self.long_pads = long_pads
        self.pin_cnt = pin_cnt

    @L.rt_field
    def pins(self):
        return times(self.pin_cnt, F.Pad)

    @L.rt_field
    def kicad_footprint(self):
        class _has_kicad_footprint(F.has_kicad_footprint_equal_ifs):
            @staticmethod
            def get_kicad_footprint() -> str:
                return "Package_DIP:DIP-{leads}_W{spacing:.2f}mm{longpads}".format(
                    leads=len(self.pins),
                    spacing=self.spacing.to(P.mm).m,
                    longpads="_LongPads" if self.long_pads else "",
                )

        return _has_kicad_footprint()

    equal_pins_in_ifs: F.has_equal_pins_in_ifs
    attach_via_pinmap: F.can_attach_via_pinmap_equal



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import os

logger = logging.getLogger(__name__)


def in_test() -> bool:
    return "PYTEST_CURRENT_TEST" in os.environ



from typing import TYPE_CHECKING

from atopile.cli.logging_ import LoggingStage
from atopile.config import BuildType, config
from faebryk.library import _F as F

if TYPE_CHECKING:
    from faebryk.core.module import Module


def init_app() -> "Module":
    with LoggingStage(name=f"init-{config.build.name}", description="Initializing app"):
        match config.build.build_type:
            case BuildType.ATO:
                return _init_ato_app()
            case BuildType.PYTHON:
                app = _init_python_app()
                app.add(F.is_app_root())
                return app
            case _:
                raise ValueError(f"Unknown build type: {config.build.build_type}")


def _init_python_app() -> "Module":
    """Initialize a specific .py build."""

    from atopile import errors
    from faebryk.libs.util import import_from_path

    try:
        app_class = import_from_path(
            config.build.entry_file_path, config.build.entry_section
        )
    except FileNotFoundError as e:
        raise errors.UserFileNotFoundError(
            f"Cannot find build entry {config.build.address}"
        ) from e
    except Exception as e:
        raise errors.UserPythonModuleError(
            f"Cannot import build entry {config.build.address}"
        ) from e

    if not isinstance(app_class, type):
        raise errors.UserPythonLoadError(
            f"Build entry {config.build.address} is not a module we can instantiate"
        )

    try:
        app = app_class()
    except Exception as e:
        raise errors.UserPythonConstructionError(
            f"Cannot construct build entry {config.build.address}"
        ) from e

    return app


def _init_ato_app() -> "Module":
    """Initialize a specific .ato build."""

    from atopile import front_end
    from atopile.datatypes import TypeRef
    from faebryk.libs.library import L

    node = front_end.bob.build_file(
        config.build.entry_file_path,
        TypeRef.from_path_str(config.build.entry_section),
    )
    assert isinstance(node, L.Module)
    return node
